from manim import *
from .utils import angle_between_vectors_signed
from .round_corners import *
from .path_mapper import *

class Pointer_Label_Free(VDict):
    def __init__(self,
                 point,
                 text:str,
                 offset_vector=(RIGHT+DOWN),
                 pointer_offset=0,
                 **kwargs):
        text_buff = 0.1
        if not 'stroke_width' in kwargs:
            kwargs['stroke_width'] = DEFAULT_STROKE_WIDTH
        super().__init__(**kwargs)

        if isinstance(text,str):
            textmob = Text(text,**kwargs)
        elif isinstance(text,Mobject):
            textmob = text
        else:
            textmob = Text('A',**kwargs)
        self.add({'text': textmob})
        self.twidth = (self['text'].get_critical_point(RIGHT)-self['text'].get_critical_point(LEFT))[0]
        self.twidth = self.twidth + text_buff * 2

        if 'stroke_width' in kwargs:
            stroke_width_loc = kwargs['stroke_width']
        else:
            stroke_width_loc = DEFAULT_STROKE_WIDTH

        self.pointer_offset = pointer_offset + stroke_width_loc/100

        dim_line = VMobject(**kwargs).set_points_as_corners([point + normalize(offset_vector)*self.pointer_offset,
                                                             point + offset_vector,
                                                             point+offset_vector +
                                                             self.twidth*np.sign(offset_vector[0])*RIGHT])
        self.add({'line':dim_line})
        self.add({'arrow':CAD_ArrowHead(self['line'],anchor_point=0,**kwargs)})
        self['arrow'].arrowhead.rotate(PI,about_point=ORIGIN)
        self['arrow'].add_updater(self['arrow'].default_updater)
        self['arrow'].update()
        theight = (self['text'].get_critical_point(UP)-self['text'].get_critical_point(DOWN))[1]
        self['text'].move_to(self['line'].points[3,:]+UP*theight*0.75,aligned_edge=LEFT*np.sign(offset_vector[0]))




    def update_point(self, point, offset_vector=(RIGHT+DOWN)):
        self['line'].set_points_as_corners([point + normalize(offset_vector)*self.pointer_offset,
                                            point + offset_vector,
                                            point + offset_vector +
                                            self.twidth * np.sign(offset_vector[0]) * RIGHT])

        theight = (self['text'].get_critical_point(UP)-self['text'].get_critical_point(DOWN))[1]
        self['text'].move_to(self['line'].points[3,:]+UP*theight*0.75,aligned_edge=LEFT*np.sign(offset_vector[0]))


class Pointer_To_Mob(Pointer_Label_Free):
    def __init__(self,
                 mob:Mobject,
                 proportion,
                 text:str,
                 dist=1,
                 pointer_offset=0,
                 **kwargs):
        point = mob.point_from_proportion(proportion)

        # if the mob center and point happens to be the same, it causes problems
        # it can happen if all the mob is 1 point
        offset_ref = point - mob.get_center()
        if 'offset' in kwargs:
            offset=kwargs['offset']
            kwargs.pop('offset')
        else:
            if np.linalg.norm(offset_ref)>1e-6:
                offset = normalize(point - mob.get_center())*dist
            else:
                # I had no better idea to handle this than to go upright
                offset = normalize(RIGHT+UP)*dist
        super().__init__(point,text, offset_vector=offset, pointer_offset=pointer_offset, **kwargs)

    def update_mob(self,mob, proportion, dist=1):
        point = mob.point_from_proportion(proportion)
        # if the mob center and point happens to be the same, it causes problems
        # it can happen if all the mob is 1 point
        offset_ref = point - mob.get_center()
        if np.linalg.norm(offset_ref) > 1e-6:
            offset = normalize(point - mob.get_center()) * dist
        else:
            # I had no better idea to handle this than to go upright
            offset = normalize(RIGHT + UP) * dist
        super().update_point(point,offset)


class Linear_Dimension(VDict):
    def __init__(self,
                 start,
                 end,
                 text=None,
                 direction=ORIGIN,
                 outside_arrow=False,
                 offset=2,
                 ext_line_offset=0,
                 tip_len=DEFAULT_ARROW_TIP_LENGTH,
                 **kwargs):
        super().__init__(**kwargs)
        self.start = start
        self.end = end
        diff_vect = end-start
        norm_vect = normalize(rotate_vector(diff_vect,PI/2))


        if direction is ORIGIN:
            ofs_vect = norm_vect * offset
            ofs_dir = norm_vect
        else:
            ofs_dir = normalize(direction)
            ofs_vect = ofs_dir * offset
        if not 'stroke_width' in kwargs:
            kwargs['stroke_width'] = DEFAULT_STROKE_WIDTH

        startpoint = start + ofs_dir * np.dot((diff_vect), ofs_dir) / 2 + ofs_vect
        endpoint = end - ofs_dir * np.dot((diff_vect), ofs_dir) / 2 + ofs_vect

        self.arrow_offset = 0
        if not outside_arrow:
            if 'stroke_width' in kwargs:
                self.arrow_offset = kwargs['stroke_width'] * 1 / 100
            else:
                self.arrow_offset = DEFAULT_STROKE_WIDTH*1/100
            ext_dir = normalize(endpoint-startpoint)
            main_line = Line(start=startpoint + self.arrow_offset*ext_dir,
                             end=endpoint - self.arrow_offset*ext_dir,
                             **kwargs)
            arrow1 = CAD_ArrowHead(main_line,anchor_point=1, arrow_size=tip_len, reversed_arrow=False)
            arrow2 = CAD_ArrowHead(main_line, anchor_point=0, arrow_size=tip_len, reversed_arrow=True)
        else:
            extension = tip_len*3*(normalize(endpoint-startpoint))
            main_line = Line(start=startpoint-extension,
                             end=endpoint+extension,
                             **kwargs)
            arrow1 = CAD_ArrowHead(main_line, anchor_point=1, arrow_size=tip_len, reversed_arrow=True)
            arrow1.arrowhead.shift(-np.linalg.norm(extension)*RIGHT)
            arrow1.default_updater(0)
            arrow2 = CAD_ArrowHead(main_line, anchor_point=0, arrow_size=tip_len, reversed_arrow=False)
            arrow2.arrowhead.shift(np.linalg.norm(extension) * RIGHT)
            arrow2.default_updater(0)
            arrow1.update()
            arrow2.update()



        self.add({'ext_line_1': Line(start=start + ofs_dir * ext_line_offset,
                                     end=startpoint + 0.25 * (normalize(startpoint-start)),
                                     **kwargs)})
        self.add({'ext_line_2': Line(start=end + ofs_dir * ext_line_offset,
                                     end=endpoint + 0.25 * (normalize(endpoint-end)),
                                     **kwargs)})
        self.add({'main_line': main_line})
        self.add({'arrow1': arrow1})
        self.add({'arrow2': arrow2})

        if isinstance(text, str):
            textmob = Text(text, **kwargs)
            textmob.set_stroke(opacity=0)
        elif isinstance(text, Mobject):
            textmob = text
        else:
            dist = np.linalg.norm(startpoint-endpoint)
            textmob = Text(f"{dist:.2}",**kwargs)
            textmob.set_stroke(opacity=0)

        text_angle = (main_line.get_angle()+PI/2)%PI-PI/2
        if abs(text_angle+PI/2)<1e-8:
            text_angle=PI/2
        self.text_h = textmob.height
        text_w = textmob.width
        if not outside_arrow:
            text_space = np.linalg.norm(self.start - self.end)-tip_len*2
        else:
            text_space = np.linalg.norm(self.start - self.end)


        if text_w > (text_space*0.8):
            textmob.scale((text_space*0.8)/text_w)
            self.text_h = textmob.height

        textmob.rotate(text_angle)
        textmob.move_to(self['main_line'].get_center() + rotate_vector(UP,text_angle)*self.text_h)
        self.add({'text': textmob})


class Angle_Dimension_3point(VDict):
    def __init__(self,
                 start,
                 end,
                 arc_center,
                 offset=2,
                 text=None,
                 outside_arrow=False,
                 ext_line_offset=0,
                 tip_len=DEFAULT_ARROW_TIP_LENGTH,
                 **kwargs):
        super().__init__(**kwargs)

        if not 'stroke_width' in kwargs:
            kwargs['stroke_width'] = DEFAULT_STROKE_WIDTH

        self.angle = angle_between_vectors_signed(start-arc_center,end-arc_center)
        radius = (np.linalg.norm(start-arc_center)+np.linalg.norm(end-arc_center))/2 + offset
        angle_0 = angle_of_vector(start-arc_center)
        angle_1 = angle_between_vectors_signed(start-arc_center,end-arc_center)

        base_arc = Arc(radius=radius,
                       start_angle=angle_0,
                       arc_center=arc_center,
                       angle=angle_1,
                       **kwargs)
        arc_p0 = base_arc.point_from_proportion(0)
        arc_p1 = base_arc.point_from_proportion(1)
        line1 = Line(start=start + normalize(arc_p0-start) * ext_line_offset,
                     end=arc_p0 + normalize(arc_p0-start)*tip_len,
                     **kwargs
        )
        line2 = Line(start=end + normalize(arc_p1-end) * ext_line_offset,
                     end=arc_p1 + normalize(arc_p1-end)*tip_len,
                     **kwargs
                     )
        # self.add(line1,line2)
        self.add({'ext_line_1': line1})
        self.add({'ext_line_2': line2})
        if not outside_arrow:

            if 'stroke_width' in kwargs:
                self.arrow_offset = kwargs['stroke_width'] * 1 / 100
            else:
                self.arrow_offset = DEFAULT_STROKE_WIDTH*1/100

            arrow1 = CAD_ArrowHead(base_arc, anchor_point=1, arrow_size=tip_len, reversed_arrow=False)
            arrow2 = CAD_ArrowHead(base_arc, anchor_point=0, arrow_size=tip_len, reversed_arrow=True)

            arrow1.arrowhead.shift(LEFT*self.arrow_offset)
            arrow2.arrowhead.shift(RIGHT * self.arrow_offset)
            arrow1.generate_points()
            arrow2.generate_points()
            self.add({'base_arc': base_arc})
            self.add({'arrow_1': arrow1})
            self.add({'arrow_2': arrow2})
        else:
            extension = tip_len*3 * np.sign(angle_1)
            angle_ext = extension/radius

            base_arc = Arc(radius=radius,
                           start_angle=angle_0-angle_ext,
                           angle=+self.angle + angle_ext * 2,
                           arc_center=arc_center,
                           **kwargs)
            arrow1 = CAD_ArrowHead(base_arc, anchor_point=1, arrow_size=tip_len, reversed_arrow=True)
            arrow1.arrowhead.shift(extension * RIGHT)
            arrow1.default_updater(0)
            arrow2 = CAD_ArrowHead(base_arc, anchor_point=0, arrow_size=tip_len, reversed_arrow=False)
            arrow2.arrowhead.shift(-extension * RIGHT)
            arrow2.default_updater(0)
            self.add({'base_arc': base_arc})
            self.add({'arrow_1': arrow1})
            self.add({'arrow_2': arrow2})

        if isinstance(text,str):
            textmob = Text(text,**kwargs)
            textmob.set_stroke(opacity=0)
        elif isinstance(text,Mobject):
            textmob = text
        else:
            textmob = Text(f"{abs(angle_1/DEGREES):.0f}°", **kwargs)
            textmob.set_stroke(opacity=0)

        pos_text = base_arc.point_from_proportion(0.5)
        angle_text = (angle_of_vector(base_arc.point_from_proportion(0.5+1e-6) -
                                      (base_arc.point_from_proportion(0.5-1e-6))) + PI / 2) % PI - PI / 2
        if abs(angle_text+PI/2)<1e-8:
            angle_text=PI/2
        self.text_h = textmob.height
        textmob.rotate(angle_text)
        textmob.move_to(pos_text + rotate_vector(UP,angle_text)*self.text_h)
        self.add({'text': textmob})

class Angle_Dimension_Mob(Angle_Dimension_3point):
    def __init__(self,
                 Mob: Mobject,
                 start_proportion,
                 end_proportion,
                 **kwargs):
        start = Mob.point_from_proportion(start_proportion)
        end = Mob.point_from_proportion(end_proportion)
        start_diff_points = np.clip(np.array([start_proportion+1e-6,start_proportion-1e-6]),0,1)
        end_diff_points = np.clip(np.array([end_proportion + 1e-6, end_proportion - 1e-6]),0,1)
        v_start = normalize(Mob.point_from_proportion(start_diff_points[1])-Mob.point_from_proportion(start_diff_points[0]))
        v_end = normalize(Mob.point_from_proportion(end_diff_points[1])-Mob.point_from_proportion(end_diff_points[0]))
        v_mat = np.concatenate((np.reshape(v_start[0:2], (2,1)),np.reshape(-v_end[0:2], (2, 1))),axis=1)
        dP = end-start
        k = np.linalg.solve(v_mat,dP[0:2])
        center = start+k[0]*v_start
        # center = np.reshape(center,(3,0))
        super().__init__(start,
                         end,
                         center.flatten(),
                         **kwargs)

class CAD_ArrowHead(Curve_Warp):
    def __init__(self,
                 target_curve: VMobject,
                 anchor_point=1,
                 arrow_size=DEFAULT_ARROW_TIP_LENGTH,
                 reversed_arrow=False,
                 **kwargs):

        # set default value
        self.__reversed_arrow = False
        self.arrowhead = VMobject()
        self.arrowhead.set_points_as_corners([(LEFT*2+UP),
                                              ORIGIN,
                                              LEFT*2+DOWN])


        # set input value
        self.reversed_arrow = reversed_arrow
        self.set_arrow_size(arrow_size)
        self.arrowhead.match_style(target_curve)

        super().__init__(self.arrowhead,target_curve,anchor_point=anchor_point,**kwargs)


    def default_updater(self,mob):
        self.PM.generate_length_map()
        self.generate_points()

    def get_tip_pos(self):
        ''' By convention, the tip of the arrow should be the right-most point of the shape, unless it's reversed'''
        if self.__reversed_arrow:
            return self.arrowhead.get_critical_point(LEFT)[0]
        else:
            return self.arrowhead.get_critical_point(RIGHT)[0]

    def set_arrow_size(self,arrow_size):
        act_size = (self.arrowhead.get_critical_point(RIGHT)-self.arrowhead.get_critical_point(LEFT))[0]
        if act_size!=0:
            self.arrowhead.shift(-self.get_tip_pos()*RIGHT)
            self.arrowhead.points *= arrow_size/act_size
            self.arrowhead.shift(+self.get_tip_pos() * RIGHT)

    def flip_arrow(self):
        self.arrowhead.flip()

    @property
    def reversed_arrow(self):
        return self.__reversed_arrow

    @reversed_arrow.setter
    def reversed_arrow(self, rev):
        if rev != self.__reversed_arrow:
            self.flip_arrow()
        self.__reversed_arrow = rev



import numpy as np
from manim import *
from scipy.optimize import fsolve
from scipy.optimize import root
from scipy.optimize import root_scalar
from .utils import angle_between_vectors_signed

__all__ = [
    "Round_Corner_Param",
    "Round_Corners",
    "Chamfer_Corner_Param",
    "Chamfer_Corners"
]


def Round_Corner_Param(radius,curve_points_1,curve_points_2):
    bez_func_1 = bezier(curve_points_1)
    diff_func_1 = bezier((curve_points_1[1:, :] - curve_points_1[:-1, :]) / 3)
    bez_func_2 = bezier(curve_points_2)
    diff_func_2= bezier((curve_points_2[1:, :] - curve_points_2[:-1, :]) / 3)

    def find_crossing(p1,p2,n1,n2):
        t = fsolve(lambda t: p1[:2]+n1[:2]*t[0]-(p2[:2]+n2[:2]*t[1]),[0,0])
        return t, p1+n1*t[0]

    def rad_cost_func(t):
        angle_sign = np.sign( angle_between_vectors_signed(diff_func_1(t[0]),diff_func_2(t[1])))
        p1 = bez_func_1((t[0]))
        n1 = normalize(rotate_vector(diff_func_1(t[0]),angle_sign* PI / 2))
        p2 = bez_func_2((t[1]))
        n2 = normalize(rotate_vector(diff_func_2(t[1]), angle_sign* PI / 2))
        d = (find_crossing(p1, p2, n1, n2))[0]
        # 2 objectives for optimization:
        #  - the normal distances should be equal to each other
        #  - the normal distances should be equal to the target radius
        # I'm hoping that in this form at least a tangent circle will be found (first goal),
        # even if there is no solution at the desired radius. I don't really know, fsolve() and roots() is magic.
        return ((d[0])-(d[1])),((d[1])+(d[0])-2*radius)

    k = root(rad_cost_func,np.asarray([0.5,0.5]),method='hybr')['x']

    p1 = bez_func_1(k[0])
    n1 = normalize(rotate_vector(diff_func_1(k[0]), PI / 2))
    p2 = bez_func_2(k[1])
    n2 = normalize(rotate_vector(diff_func_2(k[1]), PI / 2))
    d, center = find_crossing(p1, p2, n1, n2)
    r = abs(d[0])
    start_angle = np.arctan2((p1-center)[1],(p1-center)[0])
    cval = np.dot(p1-center,p2-center)
    sval = (np.cross(p1-center,p2-center))[2]
    angle = np.arctan2(sval,cval)

    out_param = {'radius': r, 'arc_center': center, 'start_angle': start_angle, 'angle': angle}

    return out_param, k


def Round_Corners(mob:VMobject,radius=0.2):
    i=0
    while i < mob.get_num_curves() and i<1e5:
        ind1 = i % mob.get_num_curves()
        ind2 = (i+1) % mob.get_num_curves()
        curve_1 = mob.get_nth_curve_points(ind1)
        curve_2 = mob.get_nth_curve_points(ind2)
        handle1 = curve_1[-1,:]-curve_1[-2,:]
        handle2 = curve_2[1, :] - curve_2[0, :]
        # angle_test = (np.cross(normalize(anchor1),normalize(anchor2)))[2]
        angle_test = angle_between_vectors_signed(handle1,handle2)
        if abs(angle_test)>1E-6:
            params, k = Round_Corner_Param(radius,curve_1,curve_2)
            cut_curve_points_1 = partial_bezier_points(curve_1, 0, k[0])
            cut_curve_points_2 = partial_bezier_points(curve_2, k[1], 1)
            loc_arc = Arc(**params,num_components=5)
            # mob.points = np.delete(mob.points, slice((ind1 * 4), (ind1 + 1) * 4), axis=0)
            # mob.points = np.delete(mob.points, slice((ind2 * 4), (ind2 + 1) * 4), axis=0)
            mob.points[ind1 * 4:(ind1 + 1) * 4, :] = cut_curve_points_1
            mob.points[ind2 * 4:(ind2 + 1) * 4, :] = cut_curve_points_2
            mob.points = np.insert(mob.points,ind2*4,loc_arc.points,axis=0)
            i=i+loc_arc.get_num_curves()+1
        else:
            i=i+1

        if i==mob.get_num_curves()-1 and not mob.is_closed():
            break

    return mob



def Chamfer_Corner_Param(offset,curve_points_1,curve_points_2):
    # this is ugly, I know, don't judge
    if hasattr(offset,'iter'):
        ofs = [offset[0], offset[1]]
    else:
        ofs = [offset, offset]
    bez_func_1 = bezier(curve_points_1)
    bez_func_2 = bezier(curve_points_2)

    #copied from vectorized mobject length stuff
    def get_norms_and_refs(curve):
        sample_points = 10
        refs = np.linspace(0, 1, sample_points)
        points = np.array([curve(a) for a in np.linspace(0, 1, sample_points)])
        diffs = points[1:] - points[:-1]
        norms = np.cumsum(np.apply_along_axis(np.linalg.norm, 1, diffs))
        norms = np.insert(norms,0,0)
        return norms,refs

    norms1,refs1 = get_norms_and_refs(bez_func_1)
    norms2,refs2 = get_norms_and_refs(bez_func_2)
    a1 = (np.interp(norms1[-1]-ofs[0], norms1, refs1))
    a2 = (np.interp(ofs[1], norms2, refs2))
    p1 = bez_func_1(a1)
    p2 = bez_func_2(a2)
    param = {'start':p1,'end':p2}

    return param, [a1,a2]


def Chamfer_Corners(mob:VMobject,offset=0.2):
    i=0
    while i < mob.get_num_curves() and i<1e5:
        ind1 = i % mob.get_num_curves()
        ind2 = (i+1) % mob.get_num_curves()
        curve_1 = mob.get_nth_curve_points(ind1)
        curve_2 = mob.get_nth_curve_points(ind2)
        handle1 = curve_1[-1,:]-curve_1[-2,:]
        handle2 = curve_2[1, :] - curve_2[0, :]
        # angle_test = (np.cross(normalize(anchor1),normalize(anchor2)))[2]
        angle_test = angle_between_vectors_signed(handle1,handle2)
        if abs(angle_test)>1E-6:
            params, k = Chamfer_Corner_Param(offset,curve_1,curve_2)
            cut_curve_points_1 = partial_bezier_points(curve_1, 0, k[0])
            cut_curve_points_2 = partial_bezier_points(curve_2, k[1], 1)
            loc_line = Line(**params)
            # mob.points = np.delete(mob.points, slice((ind1 * 4), (ind1 + 1) * 4), axis=0)
            # mob.points = np.delete(mob.points, slice((ind2 * 4), (ind2 + 1) * 4), axis=0)
            mob.points[ind1 * 4:(ind1 + 1) * 4, :] = cut_curve_points_1
            mob.points[ind2 * 4:(ind2 + 1) * 4, :] = cut_curve_points_2
            mob.points = np.insert(mob.points,ind2*4,loc_line.points,axis=0)
            i=i+loc_line.get_num_curves()+1
        else:
            i=i+1

        if i==mob.get_num_curves()-1 and not mob.is_closed():
            break

    return mob





# with tempconfig({"quality": "medium_quality", "disable_caching": True}):
#     scene = Test_corners()
#     scene.render()


import numpy as np
from manim import *

__all__ = ["angle_between_vectors_signed",
           "Bezier_Handlebars"]

def angle_between_vectors_signed(v1,v2):
    '''
    Get signed angle between vectors according to right hand rule.
    :param v1: first vector
    :param v2: second vector
    :return: angle of rotation that rotates v1 to be co-linear with v2. Range: -PI...+PI
    '''
    cval = np.dot(v1, v2)
    sval = (np.cross(v1, v2))[2]
    return np.arctan2(sval, cval)

class Bezier_Handlebars(VDict):
    '''
    Creates circles (dots) and lines on the points of another mobject.
    Used for visualizing bezier control points and handles, helps debugging.
    '''
    def __init__(self,target_mobject:VMobject, **kwargs):
        super().__init__(**kwargs)
        self.target = target_mobject
        self.generate_circles(0.03)
        self.generate_lines()


    def generate_circles(self,r):
        DotGroup = VGroup(*[
            Circle(radius=r,
                   arc_center=p,
                   color=TEAL,
                   fill_opacity=1,
                   stroke_opacity=0,
                   num_components=4)
            for p in self.target.points])
        self['dots'] = DotGroup

    def move_circles(self):
        for zipthing in zip(self['dots'],self.target.points):
            zipthing[0].move_to(zipthing[1])

    def generate_lines(self):
        pointlist = zip(self.target.points[0::4, :], self.target.points[1::4, :], self.target.points[2::4, :],
                        self.target.points[3::4, :])
        LineGroup = VGroup(*[VGroup(*[
            Line(p[0],p[1],color=TEAL, stroke_width=2),Line(p[2],p[3],color=TEAL, stroke_width=2)]
                                    ) for p in pointlist])
        self['lines'] = LineGroup

    def move_lines(self):
        pointlist = zip(self.target.points[0::4, :], self.target.points[1::4, :], self.target.points[2::4, :],
                        self.target.points[3::4, :])
        for zipthing in zip(self['lines'],pointlist):
            # Line.put_start_and_end_on()
            zipthing[0][0].put_start_and_end_on(zipthing[1][0],zipthing[1][1])
            zipthing[0][1].put_start_and_end_on(zipthing[1][2], zipthing[1][3])


import numpy as np
from manim import *
from scipy.optimize import root

class Hatch_lines(VGroup):
    def __init__(self, target_mobject: Mobject, angle=PI/6, offset=0.3,**kwargs):
        super().__init__(**kwargs)
        self.target = target_mobject
        target_size_xy = [self.target.get_critical_point(RIGHT)-self.target.get_critical_point(LEFT),
                       self.target.get_critical_point(UP)-self.target.get_critical_point(DOWN)]
        target_size_diag = np.linalg.norm(target_size_xy)
        num_lines = int(target_size_diag//offset)
        line_v = np.array([np.cos(angle),np.sin(angle),0])
        offs_v = np.array([-np.sin(angle),np.cos(angle),0])
        center = self.target.get_center()
        for k in range(num_lines*2):
            line_loc = Line(center-line_v*target_size_diag*1.5,center+line_v*target_size_diag*1.5)
            line_loc.shift((k-(num_lines-1))*offs_v*offset)
            # line_loc = Intersection(line_loc,self.target)
            intersect_idx = curve_intersection(self.target,line_loc)
            if any(intersect_idx[1]):

                intersect_indexes = sorted(intersect_idx[1])
                for j in range(len(intersect_idx[1])//2):
                    idx1 = int(intersect_indexes[0+2*j] // 1)
                    alpha1 =   intersect_indexes[0+2*j] % 1
                    idx2 = int(intersect_indexes[1+2*j] // 1)
                    alpha2 =   intersect_indexes[1+2*j] % 1
                    point1 = line_loc.get_nth_curve_function(idx1)(alpha1)
                    point2 = line_loc.get_nth_curve_function(idx2)(alpha2)
                    line_loc2 = Line(start=point1,end=point2,**kwargs)
                    self.add(line_loc2)


def curve_intersection(vmob1: VMobject, vmob2: VMobject):
    """Intersection points of 2 VMobjects.
    Finds all intersections of 2 vmobjects, as long as there is only 1 interesection per bezier subcurve.
    Self intersections are not found."""

    intersect_indx_1 = np.array([])
    intersect_indx_2 = np.array([])
    for i in range(vmob1.get_num_curves()):
        for j in range(vmob2.get_num_curves()):
            curve_1 = vmob1.get_nth_curve_points(i)
            curve_2 = vmob2.get_nth_curve_points(j)
            x_range_1 = np.array([np.amax(curve_1[:, 0]), np.amin(curve_1[:, 0])])
            x_range_2 = np.array([np.amax(curve_2[:, 0]), np.amin(curve_2[:, 0])])
            y_range_1 = np.array([np.amax(curve_1[:, 1]), np.amin(curve_1[:, 1])])
            y_range_2 = np.array([np.amax(curve_2[:, 1]), np.amin(curve_2[:, 1])])

            distinct_x = x_range_2[1] > x_range_1[0] or x_range_1[1] > x_range_2[0]
            distinct_y = y_range_2[1] > y_range_1[0] or y_range_1[1] > y_range_2[0]

            overlap = not (distinct_x or distinct_y)

            if overlap:
                curve_fun_1 = vmob1.get_nth_curve_function(i)
                curve_fun_2 = vmob2.get_nth_curve_function(j)

                sol = root(lambda t: (curve_fun_1(t[0])[0:2]-curve_fun_2(t[1])[0:2]),np.array((0.5,0.5)))
                if sol.success:
                    if 0 < sol.x[0] < 1 and 0 < sol.x[1] < 1:
                        intersect_indx_1 = np.append(intersect_indx_1, sol.x[0] + i)
                        intersect_indx_2 = np.append(intersect_indx_2, sol.x[1] + j)

    return intersect_indx_1, intersect_indx_2






import numpy as np
from manim import *
from .utils import angle_between_vectors_signed
from .round_corners import *

class Path_mapper(VMobject):
    def __init__(self,path_source:VMobject,num_of_path_points=100,**kwargs):
        super().__init__(**kwargs)
        self.num_of_path_points = num_of_path_points
        self.path= path_source
        self.generate_length_map()

    def generate_length_map(self):
        norms = np.array(0)
        for k in range(self.path.get_num_curves()):
            norms = np.append(norms, self.path.get_nth_curve_length_pieces(k,sample_points=11))
        # add up length-pieces in array form
        self.pathdata_lengths  = np.cumsum(norms)
        self.pathdata_alpha = np.linspace(0, 1, self.pathdata_lengths.size)

    def cubic_to_quads(self, cubic_points):
        # based on https://ttnghia.github.io/pdf/QuadraticApproximation.pdf
        q_points = np.empty((0,3))
        gamma = 0.5
        q1 = cubic_points[0,:] + 3/2*gamma * (cubic_points[1,:]-cubic_points[0,:])
        q3 = cubic_points[3,:] + 3/2*(1-gamma) * (cubic_points[2,:]-cubic_points[3,:])
        q2 = (1-gamma)*q1 + gamma * q3
        q0 = cubic_points[0,:]
        q4 = cubic_points[3,:]
        q_points = np.append(q_points,(q0,q1,q2,q2,q3,q4),axis=0)
        return q_points

    def calc_len_bezier_quad(self,points):
        a = points[0, :]
        b = points[1, :]
        c = points[2, :]

        B = b-a
        F = c-b
        A = F-B
        nF = np.linalg.norm(F)
        nA = np.linalg.norm(A)
        nB = np.linalg.norm(B)
        if nA>1e-8:
            L = (nF * np.dot(A, F) - nB * np.dot(A, B)) / (nA ** 2) + (nA ** 2 * nB ** 2 - np.dot(A, B) ** 2) / \
                (nA ** 3) * (np.log(nA * nF + np.dot(A, F)) - np.log(nA * nB + np.dot(A, B)))
            return L
        else:
            return np.linalg.norm(a-c)

    def calc_len_with_quads(self):
        L = 0
        for k in range(self.path.get_num_curves()):
            points = self.path.get_nth_curve_points(k)
            quads = self.cubic_to_quads(points)
            L0 = self.calc_len_bezier_quad(quads[:3, :])
            L1 = self.calc_len_bezier_quad(quads[3:, :])
            L += L0+L1
        return L

    def get_path_length(self):
        return self.pathdata_lengths[-1]

    def alpha_from_length(self,s):
        if hasattr(s, '__iter__'):
            return [np.interp(t, self.pathdata_lengths, self.pathdata_alpha) for t in s]
        else:
            return np.interp(s, self.pathdata_lengths, self.pathdata_alpha)

    def length_from_alpha(self,a):
        if hasattr(a, '__iter__'):
            return [np.interp(t, self.pathdata_alpha, self.pathdata_lengths) for t in a]
        else:
            return np.interp(a, self.pathdata_alpha, self.pathdata_lengths)

    def equalize_alpha(self, a):
        'used for inverting the alpha behavior'
        return self.alpha_from_length(a*self.get_path_length())

    def equalize_rate_func(self, rate_func):
        '''
        Specifically made to be used with Create() animation.
        :param rate_func: rate function to be equalized
        :return: callable new rate function
        Example:
        class test_path_mapper_anim(Scene):
            def construct(self):
                mob1 = round_corners(Triangle(fill_color=TEAL,fill_opacity=0).scale(3),0.5)
                PM = Path_mapper(mob1)
                mob2 = mob1.copy()
                mob1.shift(LEFT * 2.5)
                mob2.shift(RIGHT * 2.5)

                self.play(Create(mob1,rate_func=PM.equalize_rate_func(smooth)),Create(mob2),run_time=5)
                self.wait()
        '''
        def eq_func(t:float):
            return self.equalize_alpha(rate_func(t))
        return eq_func

    def point_from_proportion(self, alpha: float) -> np.ndarray:
        '''
         Override original implementation.
         Should be the same, except it uses pre calculated length table and should be faster a bit.
        '''
        if hasattr(alpha, '__iter__'):
            values = self.alpha_from_length(alpha * self.get_path_length())
            ret = np.empty((0,3))
            for a in values:
                if a == 1:
                    index = self.path.get_num_curves() - 1
                    remainder = 1
                else:
                    index = int(a * self.path.get_num_curves() // 1)
                    remainder = (a * self.path.get_num_curves()) % 1
                p = self.path.get_nth_curve_function(index)(remainder)
                ret = np.concatenate([ret,np.reshape(p,(1,3))],axis=0)
            return ret
        else:
            a = self.alpha_from_length(alpha*self.get_path_length())
            if a==1:
                index = self.path.get_num_curves()-1
                remainder = 1
            else:
                index = int(a * self.path.get_num_curves() // 1)
                remainder = (a * self.path.get_num_curves()) % 1
            return self.path.get_nth_curve_function(index)(remainder)

    def get_length_between_points(self,b,a):
        '''
        Signed arc length between to points.
        :param b: second point
        :param a: first point
        :return: length (b-a)
        '''
        return self.length_from_alpha(b)-self.length_from_alpha(a)

    def get_length_between_points_wrapped(self,b,a):
        ''' This function wraps around the length between two points similar to atan2 method.
        Useful for closed mobjects.
        Returns distance value between -L/2...L/2 '''
        AB = self.get_length_between_points(b,a)
        L = self.get_path_length()
        return (AB%L-L/2)%L-L/2

    def get_length_between_points_tuple(self,b,a):
        ''' Function to get the 2 absolute lengths between 2 parameters on closed mobjects.
        Useful for closed mobjects.
        :returns tuple (shorter, longer)'''

        AB = abs(self.get_length_between_points(b,a))
        L = self.get_path_length()
        if AB>L/2:
            return (L - AB), AB
        else:
            return AB, (L - AB)

    def get_bezier_index_from_length(self,s):
        a = self.alpha_from_length(s)
        nc = self.path.get_num_curves()
        indx = int(a * nc // 1)
        bz_a = a * nc % 1
        if indx==nc:
            indx = nc-1
            bz_a=1
        return (indx,bz_a)

    def get_tangent_unit_vector(self,s):
        # diff_bez_points = 1/3*(self.path.points[1:,:]-self.path.points[:-1,:])
        indx, bz_a = self.get_bezier_index_from_length(s)
        points = self.path.get_nth_curve_points(indx)
        dpoints = (points[1:,:]-points[:-1,:])/3
        bzf = bezier(dpoints)
        point = bzf(bz_a)
        return normalize(point)

    def get_tangent_angle(self,s):
        tv = self.get_tangent_unit_vector(s)
        return angle_of_vector(tv)

    def get_normal_unit_vector(self,s):
        tv = self.get_tangent_unit_vector(s)
        return rotate_vector(tv,PI/2)

    def get_curvature_vector(self,s):
        indx, bz_a = self.get_bezier_index_from_length(s)
        points = self.path.get_nth_curve_points(indx)
        dpoints = (points[1:, :] - points[:-1, :]) * 3
        ddpoints = (dpoints[1:, :] - dpoints[:-1, :]) * 2
        deriv = bezier(dpoints)(bz_a)
        dderiv = bezier(ddpoints)(bz_a)
        curv = np.cross(deriv, dderiv) / (np.linalg.norm(deriv)**3)
        return curv

    def get_curvature(self,s):
        return np.linalg.norm(self.get_curvature_vector(s))

# DashedVMobject

class Dashed_line_mobject(VDict):
    def __init__(self,target_mobject:VMobject,
                 num_dashes=15,
                 dashed_ratio=0.5,
                 dash_offset=0.0,
                 **kwargs):
        super().__init__(**kwargs)
        self.path = Path_mapper(target_mobject,num_of_path_points=10*target_mobject.get_num_curves())
        # self.path.add_updater(lambda mob: mob.generate_length_map())

        dshgrp = self.generate_dash_mobjects(
            **self.generate_dash_pattern_dash_distributed(num_dashes,dash_ratio = dashed_ratio,offset=dash_offset)
        )
        self.add({'dashes':dshgrp})
        self['dashes'].match_style(target_mobject)

    def generate_dash_pattern_metric(self,dash_len,space_len, num_dashes, offset=0):
        ''' generate dash pattern in metric curve-length space'''
        period = dash_len + space_len
        n = num_dashes
        full_len = self.path.get_path_length()
        dash_starts = [(i * period + offset) for i in range(n)]
        dash_ends = [(i * period + dash_len + offset) for i in range(n)]
        k=0
        while k<len(dash_ends):
            if dash_ends[k]<0:
                dash_ends.pop(k)
                dash_starts.pop(k)
            k+=1

        k = 0
        while k < len(dash_ends):
            if dash_starts[k] > full_len:
                dash_ends.pop(k)
                dash_starts.pop(k)
            k+=1
        return {'dash_starts':dash_starts,'dash_ends':dash_ends}

    def generate_dash_pattern_dash_distributed(self,num_dashes,dash_ratio = 0.5,offset=0.0):
        full_len = self.path.get_path_length()
        period = full_len / num_dashes
        dash_len = period * dash_ratio
        space_len = period * (1-dash_ratio)
        n = num_dashes+2

        return self.generate_dash_pattern_metric(dash_len, space_len, n, offset=(offset-1)*period)

    def generate_dash_mobjects(self,dash_starts=[0],dash_ends=[1]):
        ref_mob = self.path.path
        a_list = self.path.alpha_from_length(dash_starts)
        b_list = self.path.alpha_from_length(dash_ends)
        ret=[]
        for i in range(len(dash_starts)):
            mobcopy = VMobject().match_points(ref_mob)
            ret.append(mobcopy.pointwise_become_partial(ref_mob,a_list[i],b_list[i]))
        return VGroup(*ret)


class DashDot_mobject(Dashed_line_mobject):
    def __init__(self,
                 target_mobject: VMobject,
                 num_dashes=15,
                 dashed_ratio=0.35,
                 dash_offset=0.0,
                 dot_scale=1,
                 **kwargs):
        super().__init__(target_mobject,
                         num_dashes,
                         dashed_ratio,
                         dash_offset,
                         **kwargs)
        dot = Circle(radius=target_mobject.get_stroke_width()/100*dot_scale,
                     fill_opacity=1,
                     stroke_opacity=0,
                     num_components=6,
                     fill_color=target_mobject.get_stroke_color())
        dash_marks = self.generate_dash_pattern_dash_distributed(num_dashes,dashed_ratio,dash_offset)

        full_len = self.path.get_path_length()
        if dash_marks['dash_starts'][0]<dash_marks['dash_ends'][0]:
            dot_marks = np.array((np.array(dash_marks['dash_starts'])[1:] + np.array(dash_marks['dash_ends'])[:-1])) / 2
        else:
            dot_marks = np.array((np.array(dash_marks['dash_starts']) + np.array(dash_marks['dash_ends']))) / 2


        dot_marks = dot_marks%full_len
        dots=VGroup(*[dot.copy().move_to(self.path.point_from_proportion(m / full_len)) for m in dot_marks])
        self['dots']=dots





class Path_Offset_Mobject(VMobject):
    def __init__(self,
                 target_mobject,
                 ofs_func,
                 ofs_func_kwargs={},
                 num_of_samples=100,
                 discontinuities=[],
                 **kwargs):
        super().__init__(**kwargs)
        self.ofs_func_kwargs = ofs_func_kwargs
        self.PM = Path_mapper(target_mobject)
        self.PM.add_updater(lambda mob: mob.generate_length_map())
        self.discontinuities = discontinuities

        self.t_range = np.linspace(0, 1, num_of_samples)
        if discontinuities != []:
            t_disc = np.sort(
                np.concatenate((np.array(self.discontinuities) + 1e-7, np.array(self.discontinuities) - 1e-7)))
            self.t_range=np.sort(np.concatenate((self.t_range,t_disc)))
        self.ofs_func = ofs_func
        self.s_scaling_factor = 1/self.PM.get_path_length()
        self.points = self.generate_offset_paths()

    # this can be useful in lambdas for updaters
    def set_ofs_function_kwargs(self,ofs_func_kwargs):
        self.ofs_func_kwargs = ofs_func_kwargs

    def generate_bezier_points(self, input_func,t_range, Smoothing=True):
        # generate bezier 4-lets with numerical difference
        out_data = []
        for k in range(len(t_range)-1):
            t = t_range[k]
            t2 = t_range[k+1]
            val1 = input_func(t)
            val2 = input_func(t2)
            p1 = val1
            p4 = val2
            if Smoothing:
                diff1 = (input_func(t + 1e-6) - input_func(t)) / 1e-6
                diff2 = (input_func(t2) - input_func(t2 - 1e-6)) / 1e-6
                p2 = val1 + diff1 * (t2 - t) / 3
                p3 = val2 - diff2 * (t2 - t) / 3
            else:
                p2 = (val1 * 2 + val2) / 3
                p3 = (val1 + val2 * 2) / 3
            out_data.append([p1,p2,p3,p4])
        return out_data


    def generate_ref_curve(self):
        self.ref_curve = VMobject()
        bez_point = self.generate_bezier_points(self.PM.point_from_proportion,self.t_range)
        for point in bez_point:
            self.ref_curve.append_points(point)
        self.ref_curve_path = Path_mapper(self.ref_curve)

    def generate_offset_func_points(self,Smoothing=True):
        points = self.generate_bezier_points(lambda t: self.ofs_func(t,**self.ofs_func_kwargs),self.t_range,Smoothing=Smoothing)
        return points

    def generate_normal_vectors(self):
        s_range = self.t_range*self.PM.get_path_length()
        # generate normal vectors from tangent angles and turning them 90°
        # the angles can be interpolated with bezier, unit normal vectors would not remain 'unit' under interpolation
        angles = self.generate_bezier_points(self.PM.get_tangent_angle,s_range)
        out = []
        for angle in angles:
            out.append([np.array([-np.sin(a),np.cos(a),0])for a in angle])
        return out

    def generate_offset_paths(self,gen_ofs_point=True, gen_ref_curve=True):
        if gen_ref_curve:
            self.generate_ref_curve()
            self.norm_vectors = self.generate_normal_vectors()
        if gen_ofs_point:
            self.ofs_points = self.generate_offset_func_points()

        n = self.ref_curve.get_num_curves()
        ofs_vectors = np.empty((n*4,3))
        for k in range(len(self.ofs_points)):
            for j in range(len(self.ofs_points[k])):
                ofs_vectors[k*4+j,:] = self.norm_vectors[k][j] * self.ofs_points[k][j]

        return self.ref_curve.points + ofs_vectors

    def default_updater(self,gen_ofs_point=True, gen_ref_curve=True):
        self.points = self.generate_offset_paths(gen_ofs_point,gen_ref_curve)


class Curve_Warp(VMobject):
    def __init__(self,warp_source:VMobject,warp_curve:VMobject,anchor_point=0.5, **kwargs):

        self.warp_curve = warp_curve
        self.warp_source = warp_source
        self.PM = Path_mapper(self.warp_curve)
        self.anchor_point = anchor_point
        super().__init__(**kwargs)
        self.match_style(warp_source)

    def generate_points(self):

        s0 = self.PM.length_from_alpha(self.anchor_point)
        x_points = self.warp_source.points[:, 0] + s0
        y_points = self.warp_source.points[:, 1]
        L = self.PM.get_path_length()

        if self.warp_curve.is_closed():
            #if the curve is closed, out-of-bound x values can wrap around to the beginning
            x_points = x_points % L
            x_curve_points = self.PM.point_from_proportion(x_points/L)
            nv = [self.PM.get_normal_unit_vector(x) for x in x_points]
            y_curve_points = np.array( [tuplie[0] * tuplie[1] for tuplie in zip(nv,y_points)])
            self.points = x_curve_points + y_curve_points
        else:
            self.points = np.empty((0,3))
            for x,y in zip(x_points,y_points):
                if 0 < x <= L:
                    p = self.PM.point_from_proportion(x / L) + self.PM.get_normal_unit_vector(x)*y
                    self.points = np.append(self.points,np.reshape(p,(1,3)),axis=0)
                elif x>L:
                    endpoint = self.PM.point_from_proportion(1)
                    tanv = self.PM.get_tangent_unit_vector(L)
                    nv = rotate_vector(tanv,PI/2)
                    x_1 = x-L
                    p = endpoint + x_1 * tanv + y * nv
                    self.points = np.append(self.points, np.reshape(p,(1,3)), axis=0)
                else:
                    startpoint = self.PM.point_from_proportion(0)
                    tanv = self.PM.get_tangent_unit_vector(0)
                    nv = rotate_vector(tanv, PI / 2)
                    x_1 = x
                    p = startpoint + x_1 * tanv + y * nv
                    self.points = np.append(self.points, np.reshape(p, (1, 3)), axis=0)
